'
'  pb_ftp.bas
'
'  32-bit PowerBasic ftp routines
'  Copyright 1999-2001
'  Don Dickinson
'  ddickinson@usinternet.com
'
'
'  The following are needed by, but not $included in this module:
'     win32api.inc            Power Basic's Win API declaraions
'
'  The following are need by and $included within
'     wsconst.inc             WinSock Constants
'     ftp_cons.inc            FTP Constants and error message routines
'     ftp_help.bas            FTP helper routines
'
'  =======================================
'  Routines available in this module
'  =======================================
'     Function ftpChangeDir      (ByVal hSocket As Long, sDir As String) As Long
'     Function ftpConnect        (sServer As String, sUser As String, sPW As String) As Long
'     Function ftpDeleteFile     (ByVal hSocket As Long, remoteFile As String) As Long
'     Function ftpCurDir         (ByVal hSocket As Long) As String
'     Function ftpGetDirList     (ByVal hSocket As Long, sServer As String, sDirs() As String) As Long
'     Function ftpGetFile        (ByVal hSocket As Long, ByVal hStatus As Long, sServer As String, _
'                                 remoteFile As String, localFile As String) As Long
'     Function ftpGetFileToMemory(ByVal hSocket As Long, ByVal hStatus As Long, sServer As String, _
'                                 remoteFile As String, fileData As String) As Long
'     Function ftpGetFileList    (ByVal hSocket As Long, sServer As String, sFiles() As String) As Long
'     Function ftpGetList        (ByVal hSocket As Long, sServer As String, sList() As String) As Long
'     Function ftpMakeDir        (ByVal hSocket as Long, sDir as String) as Long
'     Function ftpPutFile        (ByVal hSocket As Long, ByVal hStatus As Long, _
'                                 remoteFile As String, localFile As String) As Long
'     Sub      ftpQuit           (ByVal hSocket As Long)
'     Function ftpMoveOrRename   (ByVal hSocket as Long, oldFile as String, newFile as String) as Long
'     Sub      ftpSetAscii       (ByVal hSocket As Long)
'     Sub      ftpSetBinary      (ByVal hSocket As Long)
'
'     Function ftpGetLastStatusCode    as Long
'     Function ftpGetLastStatusMsg     as String
'     Function ftpTranslateErrorCode   (ByVal errorCode As Long) As String
'
'  =======================================
'  How the library works - Important Notes
'  =======================================
'
'     >  Call ftpConnect first - this connects to the server. It returns
'        a valid socket handle if successful - Pass this handle to all the
'        other calls. If it fails it returns %INVALID_SOCKET
'
'     >  Call any other commands you wish to.
'
'     >  Most functions return %True if successful or %False if not.
'        If you need more detail about why the call failed, you can
'        use ftpGetLastStatusCode to return the last status code
'        returned from the ftp server (or a dos error code if one occurred)
'        The text error message can be retrieved by calling ftpGetLastStatusMsg
'
'     >  When you're done, call ftpQuit to clean up.
'
'     >  You should never need to call ftpSetBinary or ftpSetAscii
'        as all file operations are done in binary mode. The module calls
'        ftpSetBinary as needed. These will likely not be exported from
'        the final DLL version of this library.
'
'==============================================================================

'=============================================================================
'  Required Include Files
'==============================================================================
$Include "wsconst.inc"
$Include "ftp_cons.inc"
$Include "ftp_help.bas"

Global g_pbFTPReturn As String

'=============================================================================
'  ftpParseFileName
'  This function doesn't do any communications, it simply
'  returns the file name from the formatted line as return
'  by the ftp LIST command.                                       NOT EXPORTED
'==============================================================================
Function ftpParseFileName(ByVal incoming As String) As String

   Dim i As Long
   Dim iFields As Long
   Dim sAccum As String

   If (incoming = ".") Or (incoming = "..") Then
      Function = incoming
      Exit Function
   End If

   For i = 15 To 1 Step -1
      Replace Space$(i) With " " In incoming
   Next i

   iFields = ParseCount(incoming, Chr$(32))

   If iFields < 9 Then
      Function = ""
   Else
      sAccum = ""
      For i = 9 To iFields
         sAccum = sAccum + Parse$(incoming, Chr$(32), i) + " "
      Next i
      Function = Rtrim$(sAccum)
   End If

End Function

'=============================================================================
'  ftpGetLastStatusCode
'  Returns the FTP status code for the last command sent
'  These correspond to the %FTP_ constants in ftp_cons.inc
'==============================================================================
Function ftpGetLastStatusCode Alias "ftpGetLastStatusCode" Export As Long
   Function = gLastErr
End Function

'=============================================================================
'  ftpGetLastStatusMsg
'  Returns the text message returned by the ftp server
'  during the last communication.
'==============================================================================
Function ftpGetLastStatusMsg Alias "ftpGetLastStatusMsg" Export As String
   Function = gLastMsg
End Function

'=============================================================================
'  ftpGetLastStatusMsgZ
'  Wrapper of ftpGetLastStatusMsg that returns a string pointer
'==============================================================================
Function ftpGetLastStatusMsgZ Alias "ftpGetLastStatusMsgZ" Export As Dword
   g_pbFTPReturn = ftpGetLastStatusMsg + $NUL
   Function =StrPtr(g_pbFTPReturn)
End Function

'=============================================================================
'  ftpDeleteFile
'  Deletes the file specified from the current working directory of the server
'==============================================================================
Function ftpDeleteFile Alias "ftpDeleteFile" _
                  (ByVal hSocket As Long, ByVal remoteFile As String) Export As Long

   Dim buffer As String

   Tcp Print hSocket, "DELE " + remoteFile
   Tcp Line hSocket, buffer

   set_last_status buffer

   If buffer_status(buffer) = %FTP_ACTION_COMPLETED Then
      Function = %True
   Else
      Function = %False
   End If

End Function

'=============================================================================
'  ftpDeleteFileZ
'  Deletes the file specified from the current working directory of the server
'==============================================================================
Function ftpDeleteFileZ Alias "ftpDeleteFileZ" _
                  (ByVal hSocket As Long, remoteFile As Asciiz) Export As Long
   Function = ftpDeleteFile(hSocket, Trim$(remoteFile))

End Function

'=============================================================================
'  ftpMakeDir
'  Creates a directory.
'==============================================================================
Function ftpMakeDir Alias "ftpMakeDir" (ByVal hSocket As Long, ByVal sDir As String) Export As Long

   Dim buffer As String
   Tcp Print hSocket, "MKD " + sDir
   Tcp Line hSocket, buffer
   set_last_status buffer
   If buffer_status(buffer) = %FTP_DIRECTORY_OK Then
      Function = %True
   Else
      Function = %False
   End If


End Function

'=============================================================================
'  ftpMakeDirZ
'  Same as ftpMakeDir but with asciiz parameters
'==============================================================================
Function ftpMakeDirZ Alias "ftpMakeDirZ" (ByVal hSocket As Long, zDir As Asciiz) Export As Long
   Function = ftpMakeDir(hSocket, Trim$(zDir))
End Function

'=============================================================================
'  ftpRemoveDir
'  Deletes a directory
'==============================================================================
Function ftpRemoveDir Alias "ftpRemoveDir" (ByVal hSocket As Long, ByVal sDir As String) Export As Long

   Dim buffer As String
   Tcp Print hSocket, "RMD " + sDir
   Tcp Line hSocket, buffer
   set_last_status buffer
   If buffer_status(buffer) = %FTP_ACTION_COMPLETED Then
      Function = %True
   Else
      Function = %False
   End If

End Function

'=============================================================================
'  ftpRemoveDirZ
'  same as ftpRemoveDir but with asciiz parameters
'==============================================================================
Function ftpRemoveDirZ Alias "ftpRemoveDirZ" (ByVal hSocket As Long, zDir As Asciiz) Export As Long
   Function = ftpRemoveDir(hSocket, Trim$(zDir))
End Function

'=============================================================================
'  ftpMoveOrRename
'  Renames or moves a file
'==============================================================================
Function ftpMoveOrRename Alias "ftpMoveOrRename" (ByVal hSocket As Long, _
               ByVal oldFile As String, ByVal newFile As String) Export As Long

   Dim buffer As String

   '- Rename From
   Tcp Print hSocket, "RNFR " + oldFile
   Tcp Line hSocket, buffer
   If buffer_status(buffer) = %FTP_ACTION_PENDING Then

      '- Rename To
      Tcp Print hSocket, "RNTO " + newFile
      Tcp Line hSocket, buffer
      If buffer_status(buffer) = %FTP_ACTION_COMPLETED Then
         Function = %True
      Else
         Function = %False
      End If
   Else
      Function = %False
   End If
   set_last_status buffer

End Function

'=============================================================================
'  ftpMoveOrRenameZ
'  same as ftpMoveOrRename except with asciiz parameters
'==============================================================================
Function ftpMoveOrRenameZ Alias "ftpMoveOrRenameZ" (ByVal hSocket As Long, _
         oldFile As Asciiz, newFile As Asciiz) Export As Long
   Function = ftpMoveOrRename(hSocket, Trim$(oldFile), Trim$(newFile))
End Function

'=============================================================================
'  ftpSetAscii
'  turns on ascii transfer mode
'==============================================================================
Sub ftpSetAscii Alias "ftpSetAscii" (ByVal hSocket As Long) Export

   Dim buffer As String
   Tcp Print hSocket, "TYPE A"
   Tcp Line hSocket, buffer
   set_last_status buffer

End Sub

'=============================================================================
'  ftpSetBinary
'  turns on binary transfer mode
'==============================================================================
Sub ftpSetBinary Alias "ftpSetBinary" (ByVal hSocket As Long) Export

   Dim buffer As String
   Tcp Print hSocket, "TYPE I"
   Tcp Line hSocket, buffer
   set_last_status buffer

End Sub

'=============================================================================
'  ftpCurDir
'  returns the current working directory on the server
'  returns "" if an error occurs
'==============================================================================
Function ftpCurDir Alias "ftpCurDir" (ByVal hSocket As Long) Export As String

   Dim buffer As String

   Tcp Print hSocket, "PWD"
   Tcp Line hSocket, buffer
   set_last_status buffer
   If buffer_status(buffer) = 257 Then
      Function = Parse$(buffer, Chr$(34), 2)
   Else
      Function = ""
   End If

End Function

'=============================================================================
'  ftpCurDirZ
'  same as ftpCurDir except it returns a string pointer
'==============================================================================
Function ftpCurDirZ Alias "ftpCurDirZ" (ByVal hSocket As Long) Export As Dword

   g_pbFTPReturn = ftpCurDir(hSocket) + $NUL
   Function =StrPtr(g_pbFTPReturn)

End Function


'=============================================================================
'  ftpChangeDir
'  Returns %True if it can change the directory
'  Returns %False if not.
'==============================================================================
Function ftpChangeDir Alias "ftpChangeDir" (ByVal hSocket As Long, _
         ByVal sDir As String) Export As Long

   Dim buffer As String

   Tcp Print hSocket, "CWD " + Trim$(sDir)
   Tcp Line hSocket, buffer
   set_last_status buffer
   If buffer_status(buffer) = 250 Then
      Function = %True
   Else
      Function = %False
   End If

End Function

'=============================================================================
'  ftpChangeDirZ
'  same as ftpChangeDir but with PChar parameters
'==============================================================================
Function ftpChangeDirZ Alias "ftpChangeDirZ" (ByVal hSocket As Long, _
         zDir As Asciiz) Export As Long
   Function = ftpChangeDir(hSocket, Trim$(zDir))
End Function

'==============================================================================
'  ftpGetFileToMemory
'  Retrieves the file's contents and puts it in the string "incomingData"
'==============================================================================
Function ftpGetFileToMemory Alias "ftpGetFileToMemory" _
      ( ByVal hSocket As Long, _
        ByVal hStatus As Long, ByVal sServer As String, _
        ByVal remoteFile As String, incomingData As String) Export As Long

   Dim hData As Long
   Dim iProgress As Long
   Dim iTotal As Long
   Dim zText As Asciiz * 200
   Dim buffer As String

   incomingData = ""
   
   '- Make sure that we are doing a binary transfer
   ftpSetBinary hSocket

   '- Open a passive-mode connection to the server.
   hData = get_passive_handle(hSocket, sServer)
   If hData = %INVALID_SOCKET Then
      buffer = Format$(gLastErr, "000") + " " + gLastMsg
      Function = %False
   Else
      
      '- Request the file
      Tcp Print hSocket, "RETR " + remoteFile
      Tcp Line hSocket, buffer
      'If buffer_status(buffer) <> 150 Then
      If buffer_status(buffer) > 299 Then
         Function = %False
      Else

         '- The total bytes to retrieve is inside
         '  of parenthesis in the returned buffer.
         '
         iTotal = Int(Val(Parse$(buffer, Any "()", 2)))
         iProgress = 0

         '- Read in blocks until we have the whole file.
         Do
            If read_data_block(hData, buffer) < 1 Then Exit Do
            incomingData = incomingData + buffer
            iProgress = iProgress + Len(buffer)
            zText = Format$(iProgress) + " of " + Format$(iTotal)

            '- Progress is passed back via a window-handle
            If hStatus <> 0 Then
               SetWindowText hStatus, zText
            End If

         Loop
         Tcp Line hSocket, buffer
         Function = %True
      End If
      Close hData
   End If

   set_last_status buffer

End Function        

'=============================================================================
'  ftpGetFile
'  Opens a binary mode connection and downloads the requested file.
'  Returns %True if successful and %False if not.
'  If hStatus is non-zero it is assumed to be a windows handle. As the
'  status of the download changes the caption of this window is changed to
'  be the percent complete of the transfer. If the file already exists then
'  it will be erased and replaced by the downloaded file.
'  Note that this function ALWAYS operates in passive mode - that is, it will
'  always connect to the server's data socket, it doesn't create it's own.
'==============================================================================
Function ftpGetFile Alias "ftpGetFile" (ByVal hSocket As Long, _
                     ByVal hStatus As Long, ByVal sServer As String, _
                     ByVal remoteFile As String, ByVal localFile As String) Export As Long

   Dim hData As Long
   Dim iFF As Long
   Dim iProgress As Long
   Dim iTotal As Long
   Dim zText As Asciiz * 200
   Dim buffer As String

   On Error Resume Next
   iFF = FreeFile
   Kill localFile
   errclear
   If Dir$(localFile) <> "" Then
      buffer = Format$(Err, "000") + " Local File Can't be deleted"
      Function = %False
      exit function
   end if
   
   On Error Resume Next
   Open localFile For Binary Lock Read Write As #iFF
   If Err Then
      buffer = Format$(Err, "000") + " Local File Open Error"
      Function = %False
      exit function
   end if
   
   '- Make sure that we are doing a binary transfer
   ftpSetBinary hSocket

   '- Open a passive-mode connection to the server.
   hData = get_passive_handle(hSocket, sServer)
   If hData = %INVALID_SOCKET Then
      buffer = Format$(gLastErr, "000") + " " + gLastMsg
      Function = %False
   Else
      
      '- Request the file
      Tcp Print hSocket, "RETR " + remoteFile
      Tcp Line hSocket, buffer
      If buffer_status(buffer) > 299 Then
         Function = %False
      Else

         '- The total bytes to retrieve is inside
         '  of parenthesis in the returned buffer.
         '
         iTotal = Int(Val(Parse$(buffer, Any "()", 2)))
         iProgress = 0

         '- Read in blocks until we have the whole file.
         Do
            If read_data_block(hData, buffer) < 1 Then Exit Do
            Put #iFF,, buffer
            iProgress = iProgress + Len(buffer)
            zText = Format$(iProgress) + " of " + Format$(iTotal)

            '- Progress is passed back via a window-handle
            If hStatus <> 0 Then
               SetWindowText hStatus, zText
            End If

         Loop
         Tcp Line hSocket, buffer
         Function = %True
      End If
      Close hData
   End If
   Close #iFF

   set_last_status buffer

End Function

'=============================================================================
'  ftpGetFileZ
'  Same as ftpGetFile except with asciiz parameters
'==============================================================================
Function ftpGetFileZ Alias "ftpGetFileZ" (ByVal hSocket As Long, _
                     ByVal hStatus As Long, zServer As Asciiz, _
                     remoteFile As Asciiz, localFile As Asciiz) Export As Long
   Function = ftpGetFile(hSocket, hStatus, Trim$(zServer), Trim$(remoteFile), Trim$(localFile))
End Function

'=============================================================================
'  ftpPutFile
'  Opens a binary mode connection and uploads the specified file to the
'  current working directory on the server.
'  It returns %True on success or %False on failure.
'  hStatus is assumed to be a status window handle if non-zero. As the status
'  of the upload changes, the caption of this window is changed to be the
'  percent complete of the transfer
'==============================================================================
Function ftpPutFile Alias "ftpPutFile" (ByVal hSocket As Long, ByVal hStatus As Long, _
                     ByVal sServer As String, ByVal remoteFile As String, _
                     ByVal localFile As String) Export As Long

   Dim iLoop As Long
   Dim iFF As Long
   Dim iTotal As Long
   Dim iBlocks As Long
   Dim iLeft As Long
   Dim iAbort As Long
   Dim hData As Long
   Dim buffer As String
   Dim zText As Asciiz * 100

   '- If the local file doesn't exist, there's
   '  no point in continuing.
   '
   If Dir$(localFile) = "" Then
      buffer = "001 Local File Not Found"
      Function = %False
   Else

      '- Open the local file and check for error
      iFF = FreeFile
      Open localFile For Binary As #iFF
      If Err Then
         Buffer = Format$(Err, "000") + " Local File Access Error"
         Function = %False
      Else

         '- Open a data socket
         ftpSetBinary hSocket
         hData = get_passive_handle(hSocket, sServer)
         If hData = %INVALID_SOCKET Then
            buffer = Format$(gLastErr, "000") + " " + gLastMsg
            Function = %False
         Else

            '- Request the upload
            Tcp Print hSocket, "STOR " + remoteFile
            Tcp Line hSocket, buffer
            
            '- I have seen this as 150 and 125 - everything under 300 should be ok.
            If (buffer_status(buffer) < 100) or (buffer_status(buffer) > 299) Then
               Close hData
               Function = %False
            Else

               '- Send the file in chunks
               iTotal = Lof(iFF)
               iBlocks = iTotal \ %FTP_BLOCK_SIZE
               iLeft = iTotal Mod %FTP_BLOCK_SIZE
               
               iAbort = %False
               For iLoop = 1 To iBlocks
                  buffer = Space$(%FTP_BLOCK_SIZE)
                  Get #iFF,, buffer
                  If Err Then
                     iAbort = %True
                     Exit For
                  End If
                  If write_data_block(hData, buffer) = %False Then
                     iAbort = %True
                     Exit For
                  End If
                  zText = Format$(iLoop * %FTP_BLOCK_SIZE) + " of " + Format$(iTotal)
                  If hStatus > 0 Then
                     SetWindowText hStatus, zText
                  End If
                  'Tcp Print hSocket, "NOOP"
                  'Tcp Line hSocket, buffer
               Next i

               '- Send the last piece
               If (iAbort = %False) And (iLeft > 0) Then
                  buffer = Space$(iLeft)
                  Get #iFF,, buffer
                  If Err Then
                     iAbort = %True
                  Else
                     If write_data_block(hData, buffer) = %False Then
                        iAbort = %True
                     End If
                  End If
               End If
               Close hData

               '- Cleanup and return
               If iAbort Then
                  Function = %False
               Else
                  Tcp Line hSocket, buffer
                  If buffer_status(buffer) = 226 Then
                     Function = %True
                  Else
                     Buffer = Format$(buffer_status(buffer), "000") + " Unable to upload file."
                     Function = %False
                  End If
               End If
            End If

         End If
         Close #iFF
      End If
   End If

   set_last_status buffer

End Function

'=============================================================================
'  ftpPutFileZ
'  same as ftpPutFile but with asciiz parameters
'==============================================================================
Function ftpPutFileZ Alias "ftpPutFileZ" (ByVal hSocket As Long, ByVal hStatus As Long, _
                     zServer As Asciiz, remoteFile As Asciiz, _
                     localFile As Asciiz) Export As Long

   Function = ftpPutFile(hSocket, hStatus, Trim$(zServer), Trim$(remoteFile), Trim$(localFile))

End Function

'=============================================================================
'  ftpGetList
'  Fills sList() with all of the file info returned
'  by a LIST command. Returns the number of lines
'  returned. (same as ubound(sList))
'==============================================================================
Function ftpGetList Alias "ftpGetList" (ByVal hSocket As Long, _
                  ByVal sServer As String, sList() As String) Export As Long

   Dim iCount As Long
   Dim hData As Long
   Dim buffer As String

   ReDim sList(0 To 2) As String

   '- Get a passive mode socket handle
   hData = get_passive_handle(hSocket, sServer)

   If hData = %INVALID_SOCKET Then
      buffer = Format$(gLastErr, "000") + " " + gLastMsg
      Function = 0
   Else

      '- Request the directory list
      On Error Resume Next
      ftpSetAscii hSocket
      Tcp Print hSocket, "LIST"
      Tcp Line hSocket, buffer
      iCount = 2
      Do
         Tcp Line hData, buffer

         If Trim$(buffer) = "" Then
            Exit Do
         ElseIf Err Then
            Exit Do
         End If

         If (Right$(buffer, 3) <> " ..") And (Right$(buffer, 2) <> " .") Then
            iCount = iCount + 1
            ReDim Preserve sList(0 To iCount) As String
            sList(iCount) = buffer
         End If
      Loop
      sList(1) = "."
      sList(2) = ".."

      Tcp Line hSocket, buffer
      Close hData
      Function = iCount
   End If
   set_last_status buffer

End Function

'=============================================================================
'  ftpGetListZ
'  Like ftpGetList except that it returns a chr$(13) delimited list
'  of files in the zList parameter
'==============================================================================
Function ftpGetListZ Alias "ftpGetListZ" (ByVal hSocket As Long, _
                  zServer As Asciiz, zList As Dword ) Export As Long

   Dim i As Long
   Dim iLines As Long
   Dim sList() As String
   Dim sL As String

   ReDim sList(0 To 0) As String
   iLines = ftpGetList(hSocket, Trim$(zServer), sList())
   sL = ""
   For i = 1 To iLines
      If sL = "" Then
         sL = sList(i)
      Else
         sL = sL + Chr$(13) + sList(i)
      End If
   Next i
   g_pbFTPReturn = sL + $NUL
   zList = StrPtr(g_pbFTPReturn)

   Function = iLines

End Function

'=============================================================================
'  ftpGetFileList
'  Fills the array - sFiles() with a list of all the files in
'  the current directory. Directories are not included in the
'  list. The list is dimmed from 0 to the number of files in
'  the directory. The 0 element is ignored. The function returns
'  ubound(sFiles)
'==============================================================================
Function ftpGetFileList Alias "ftpGetFileList" (ByVal hSocket As Long, _
                  ByVal sServer As String, sFiles() As String) Export As Long

   Dim i As Long
   Dim iList As Long
   Dim sList() As String

   ReDim sList(0 To 0) As String
   ReDim sFiles(0 To 0) As String
   iList = ftpGetList(hSocket, sServer, sList())
   If iList > 0 Then
      For i = 1 To iList
         If Left$(sList(i), 1) = "-" Then
            ReDim Preserve sFiles(0 To UBound(sFiles) + 1) As String
            sFiles(UBound(sFiles)) = sList(i)
         End If
      Next i
   End If
   Function = UBound(sFiles)

End Function

'=============================================================================
'  ftpGetFileListZ
'  Like ftpGetFileList except it returns zFiles as a pointer
'  to a Chr$(13) delimited array of file names.
'==============================================================================
Function ftpGetFileListZ Alias "ftpGetFileListZ" (ByVal hSocket As Long, _
                  zServer As Asciiz, zFiles As Dword) Export As Long
   Dim i As Long
   Dim iCount As Long
   Dim sLines() As String
   Dim sFiles As String

   ReDim sLines(0 To 0) As String

   iCount = ftpGetFileList(hSocket, Trim$(zServer), sLines())
   sFiles = ""
   For i = 1 To iCount
      If sFiles = "" Then
         sFiles = sLines(i)
      Else
         sFiles = sFiles + Chr$(13) + sLines(i)
      End If
   Next i
   g_pbFTPReturn = sFiles + $NUL
   zFiles = StrPtr(g_pbFTPReturn)
   Function = iCount

End Function

'=============================================================================
'  ftpGetFileList2
'  Fills the array - sFiles() with a list of all the files in
'  the current directory. Directories are not included in the
'  list. The list is dimmed from 0 to the number of files in
'  the directory. The 0 element is ignored. The function returns
'  ubound(sFiles)
'  This differes from ftpGetFileList in that it only returns
'  the name of the file - no other information.
'==============================================================================
Function ftpGetFileList2 Alias "ftpGetFileList2" (ByVal hSocket As Long, _
                  ByVal sServer As String, sFiles() As String) Export As Long

   Dim iCount As Long
   Dim hData As Long
   Dim buffer As String

   ReDim sFiles(0 To 0) As String

   '- Get a passive mode socket handle
   hData = get_passive_handle(hSocket, sServer)

   If hData = %INVALID_SOCKET Then
      buffer = Format$(gLastErr, "000") + " " + gLastMsg
      Function = 0
   Else

      '- Request the directory list
      On Error Resume Next
      ftpSetAscii hSocket
      Tcp Print hSocket, "NLST"
      Tcp Line hSocket, buffer
      iCount = 2
      Do
         Tcp Line hData, buffer

         If Trim$(buffer) = "" Then
            Exit Do
         ElseIf Err Then
            Exit Do
         End If

         iCount = iCount + 1
         ReDim Preserve sFiles(0 To iCount) As String
         sFiles(iCount) = buffer
      Loop

      Tcp Line hSocket, buffer
      Close hData
      Function = iCount
   End If
   set_last_status buffer

End Function

'=============================================================================
'  ftpGetFileList2Z
'  same as ftpGetFileList except zFiles is filled with a chr$(13)
'  delimited list of file names.
'==============================================================================
Function ftpGetFileList2Z Alias "ftpGetFileList2Z" (ByVal hSocket As Long, _
                  zServer As Asciiz, zFiles As Dword) Export As Long
   Dim i As Long
   Dim iCount As Long
   Dim sLines() As String
   Dim sFiles As String

   ReDim sLines(0 To 0) As String

   iCount = ftpGetFileList2(hSocket, Trim$(zServer), sLines())
   sFiles = ""
   For i = 1 To iCount
      If sFiles = "" Then
         sFiles = sLines(i)
      Else
         sFiles = sFiles + Chr$(13) + sLines(i)
      End If
   Next i
   g_pbFTPReturn = sFiles + $NUL
   zFiles = StrPtr(g_pbFTPReturn)
   Function = iCount

End Function

'=============================================================================
'  ftpGetDirList
'  Fills the sDirs() array with a list of all the directories
'  in the current directory on the ftp server. It returns the
'  number of elements in sDirs. The array is dimensioned from
'  0 through the number of directories found. The 0 element
'  is left empty and should be ignored.
'==============================================================================
Function ftpGetDirList Alias "ftpGetDirList" _
               (ByVal hSocket As Long, ByVal sServer As String, sDirs() As String) Export As Long

   Dim i As Long
   Dim iList As Long
   Dim sList() As String

   ReDim sList(0 To 0) As String
   ReDim sDirs(0 To 0) As String
   iList = ftpGetList(hSocket, sServer, sList())
   If iList > 0 Then
      For i = 1 To iList
         If UCase$(Left$(sList(i), 1)) = "D" Then
            ReDim Preserve sDirs(0 To UBound(sDirs) + 1) As String
            sDirs(UBound(sDirs)) = sList(i)
         End If
      Next i
   End If
   Function = UBound(sDirs)

End Function

'=============================================================================
'  ftpGetDirListZ
'  Same as ftpGetDirList except
'  that it returns a chr$(13) delimited list of dirs in zDirs
'==============================================================================
Function ftpGetDirListZ Alias "ftpGetDirListZ" _
               (ByVal hSocket As Long, zServer As Asciiz, zDirs As Dword) Export As Long

   Dim i As Long
   Dim iCount As Long
   Dim sDirs As String
   Dim sD() As String

   ReDim sD(0 To 0) As String
   iCount = ftpGetDirList(hSocket, Trim$(zServer), sD())
   sDirs = ""
   For i = 1 To iCount
      If sDirs = "" Then
         sDirs = sD(i)
      Else
         sDirs = sDirs + Chr$(13) + sD(i)
      End If
   Next i

   g_pbFTPReturn = sDirs + $NUL
   zDirs = StrPtr(g_pbFTPReturn)
   Function = iCount

End Function

'=============================================================================
'  connect to ftp server
'  if successful, a valid socket handle is return
'  if not, %INVALID_SOCKET
'==============================================================================
Function ftpConnect Alias "ftpConnect" (ByVal sServer As String, _
                  ByVal sUser As String, ByVal sPW As String) Export As Long

   Dim hSocket As Long
   Dim buffer As String

   '- Connect to port
   ErrClear
   On Error Resume Next
   hSocket = FreeFile
   Tcp Open Port %IPPORT_FTP At sServer As hSocket
   If Err Then
      set_last_status Format$(Err, "000") + " Unable to open socket"
      Function = %INVALID_SOCKET
      '============
      Exit Function
      '============
   End If

   '- What does the server have to say
   Do
      Tcp Line hSocket, buffer
      If buffer_status(buffer) <> 220 Then
         set_last_status buffer
         Close hSocket
         Function = %INVALID_SOCKET
         '============
         Exit Function
         '============
      End If
      If Mid$(buffer, 4, 1) <> "-" Then Exit Do
      
   Loop

   '- Login with user and password
   Tcp Print hSocket, "USER " + sUser
   Tcp Line hSocket, buffer

   '- Send the password if requested.
   If buffer_status(buffer) = 331 Then
      Tcp Print hSocket, "PASS " + sPW
   End If

   Do
      Tcp Line hSocket, buffer
      If Mid$(buffer, 4, 1) <> "-" Then Exit Do
   Loop

   set_last_status buffer


   If buffer_status(buffer) <> 230 Then
      Close hSocket
      Function = %INVALID_SOCKET
      '============
      Exit Function
      '============
   End If

   Function = hSocket


End Function

'=============================================================================
'  ftpConnectZ
'  Same as ftpConnect except that it takes asciiz instead of strings
'  as parameters.
'==============================================================================
Function ftpConnectZ Alias "ftpConnectZ" (zServer As Asciiz, _
                  zUser As Asciiz, zPW As Asciiz) Export As Long
   Function = ftpConnect(Trim$(zServer), Trim$(zUser), Trim$(zPW))
End Function

'=============================================================================
'  ftpQuit
'  Closes an FTP connection
'==============================================================================
Sub ftpQuit Alias "ftpQuit" (ByVal hSocket As Long) Export

   Dim buffer As String

   Tcp Print hSocket, "QUIT"
   Tcp Line hSocket, buffer
   Do Until Mid$(buffer, 4, 1) <> "-"
      Tcp Line hSocket, buffer
   Loop
   set_last_status buffer
   Close hSocket
   hSocket = %INVALID_SOCKET

End Sub

'=============================================================================
'  ftpTranslateErrorCode
'  Translates an error code returned from an FTP server into
'  a common text interpretation ofthe error. It's usually better
'  to just read the string returned by the server instead of using
'  this function.
'=============================================================================
Function ftpTranslateErrorCode Alias "ftpTranslateErrorCode" (ByVal errorCode As Long) As String

   Dim e As String

   Select Case errorCode
      Case %FTP_RESTART_MARKER_REPLY
         e = "Restart Marker Reply"
      Case %FTP_SERVICE_READY_SOON
         e = "Service available soon:
      Case %FTP_OPEN_STARTING_TRANSFER
         e = "Data Connection Open. Starting Transfer"
      Case %FTP_FILE_READY
         e = "File ready. About to open data connection."
      Case %FTP_OK
         e = "Success"
      Case %FTP_COMMAND_NOT_IMPLEMENTED
         e = "Command Not Implemented"
      Case %FTP_SYSTEM_STATUS
         e = "System Status"
      Case %FTP_DIRECTORY_STATUS
         e = "Directory Status"
      Case %FTP_FILE_STATUS
         e = "File Status"
      Case %FTP_HELP_MESSAGE
         e = "Help Message"
      Case %FTP_SYSTEM_TYPE_NAME
         e = "System Type Name"
      Case %FTP_SERVICE_READY
         e = "Service Ready"
      Case %FTP_CONNECTION_CLOSING
         e = "Connection Closing. Operation Aborted"
      Case %FTP_OPEN_NO_CURRENT_TRANSFER
         e = "Connection Open. No transfer in progress."
      Case %FTP_TRANSFER_COMPLETE
         e = "Transfer completed successfully."
      Case %FTP_ENTERING_PASSIVE_MODE
         e = "Entering Passive Mode"
      Case %FTP_USER_LOGGED_IN
         e = "User Logged In"
      Case %FTP_ACTION_COMPLETED
         e = "Action Completed Successfully"
      Case %FTP_DIRECTORY_OK
         e = "Directory Created Ok or Current Directory is Ok."
      Case %FTP_PASSWORD_NEEDED
         e = "User Ok, but password needed."
      Case %FTP_ACCOUNT_NEEDED
         e = "Account needed"
      Case %FTP_ACTION_PENDING
         e = "Action Pending Further Information"
      Case %FTP_SERVICE_UNAVAILABLE
         e = "Service Unavailable"
      Case %FTP_DATA_CONN_NOT_OPENED
         e = "Data Connection Not Opened."
      Case %FTP_DATA_CONN_CLOSED
         e = "Data Connection Closed"
      Case %FTP_FILE_UNAVAILABLE
         e = "File unavailable"
      Case %FTP_LOCAL_PROCESSING_ERROR
         e = "Local Processing Error"
      Case %FTP_NOT_ENOUGH_DISK_SPACE
         e = "Not Enough Disk Space"
      Case %FTP_COMMAND_NOT_RECOGNIZED
         e = "Command Not Recognized"
      Case %FTP_BAD_PARAMETERS
         e = "Syntax Error: Invalid parameters or arguments"
      Case %FTP_NOT_IMPLEMENETED
         e = "Not Implemented"
      Case %FTP_BAD_COMMAND_SEQUENCE
         e = "Bad Command Sequence"
      Case %FTP_COMMAND_UNAVAILABLE
         e = "Command Unavailable"
      Case %FTP_USER_NOT_LOGGED_IN
         e = "User Not Logged In"
      Case %FTP_ACCOUNT_NEEDED_TO_STORE
         e = "Account Needed to Store Files"
      Case %FTP_ACCESS_DENIED
         e = "Access Denied"
      Case %FTP_PAGE_TYPE_UNKNOWN
         e = "Page Type Unknown"
      Case %FTP_STORAGE_LIMIT_EXCEEDED
         e = "Storage Limit Exceeded"
      Case %FTP_BAD_FILE_NAME
         e = "Bad File Name"
      Case else
         e = "FTP error " + Format$(errorCode)
   End Select

   Function = e

End Function

'=============================================================================
'  ftpTranslateErrorCodeZ
'  same as ftpTranslateErrorcode except that it returns a string pointer
'=============================================================================
Function ftpTranslateErrorCodeZ Alias "ftpTranslateErrorCodeZ" (ByVal errorCode As Long) As Dword
   g_pbFTPReturn = ftpTranslateErrorCode(errorCode) + $NUL
   Function = StrPtr(g_pbFTPReturn)
End Function











                                                   